/*	ToString.h - Object to String conversions.
*
*******************************************************************************
* Copyright (c) 1996-97 Digi International Inc., All Rights Reserved          *
*                                                                             *
* This  software  contains  proprietary  and  confidential   information   of *
* Digi International  Inc.   By  accepting  transfer  of this copy, Recipient *
* agrees to retain this software in  confidence,  to  prevent  disclosure  to *
* others,  and  to make no  use of this software other than that for which it *
* was delivered.  This is an unpublished copyright work  Digi   International *
* Inc.    Except   as   permitted   by  federal  law,  17 USC 117, copying is *
* strictly prohibited.                                                        *
*                                                                             *
* Digi  International  Inc.  CONFIDENTIAL  -   (Digi    International    Inc. *
* Confidential  when   combined   with   the   aggregated  modules  for  this *
* product) OBJECT CODE ONLY SOURCE MATERIALS                                  *
*                                                                             *
* (C) COPYRIGHT Digi International Inc. 1996-97                               *
*                                                                             *
* All Rights Reserved                                                         *
*                                                                             *
* US   Government   Users   Restricted   Rights   -   Use,   duplication   or *
* disclosure  restricted by GSA ADP Schedule Contract with Digi International *
* Inc.                                                                        *
*                                                                             *
*******************************************************************************
*
*/

#if !defined(__TOSTRING_H__)
#define	__TOSTRING_H__

char * CopyString(char const * const psz);
void FreeString(char /*const*/ * const psz);

/*
	These functions [*ToString()] all return a static string for speed reasons,
	if you need to call them more than once (and thereby overwriting your previous 
	call's string), use the CopyString() and FreeString() functions prototyped above.
*/

char const * SuccessToString(BOOL const fSuccess);
char const * ErrorStatusToString(AsnInteger const iaErrorStatus);
char const * ObjectTypeToString(Object_t const eObjectType);
char const * RequestToString(BYTE const eType);
char const * OidToString(AsnObjectIdentifier const * const poia);
char const * MgmtResultToString(uint32 const nResult);
char const * CacheTypeToString(uint16 const eType);
char const * MgmtCommandToString(uint16 const nCommand);
char const * UnsignedToString(unsigned n);
char const * RawBytesToString(void const * const pv, size_t const cb);

#endif

